!
!     ӥ¿༰Ÿˡ(collocationˡ)Ѥ
!     ʿ̥ݥ梁ήĹΨη׻.
!
program SolveEigenvalueProblemOfPoiseuilleFlow
  
  implicit none
  
  integer, parameter :: m = 81
  
  real(8) re, al
  real(8) y(1:m), ty(0:4, 1:m, 0:m-1), u(1:m), d2u(1:m)
  complex(8) ca(1:m, 1:m), cb(1:m, 1:m), eigenvector(1:m), eigenvalue
  
  call set_parameter(re, al)
  call create_tables(m, y, ty)
  call set_meanflow(m, y, u, d2u)
  call calculate_matrices(m, ty, u, d2u, re, al, ca, cb)
  call solve_eigenvalue_problem(m, ca, cb, eigenvector, eigenvalue)
  
  print *, 'MAXIMUM GROWTH RATE', al*eigenvalue
  
  call calculate_eigenmode(m, y, ty, eigenvector)
  
  stop
end program SolveEigenvalueProblemOfPoiseuilleFlow


! 
! ѥ᡼
! 
subroutine set_parameter(re, al)

  implicit none

  real(8), intent(out) :: re, al

  write(*,*) 'input Re ?';    read(*,*) re
  write(*,*) 'input alpha ?'; read(*,*) al

end subroutine set_parameter


!
! yεŸδؿyɸͤΥơ֥
! 
subroutine create_tables(m, y, ty)

  implicit none

  integer, intent(in)  :: m
  real(8), intent(out) :: y(1:m), &           ! ʻyɸ
       &                  ty(0:5, 1:m, 0:m-1) ! ؿ

  real(8) tn(0:4, 1:m, 0:m-1)

  call set_axis(m, y)
  call create_chebychev_polynomial_table(m, y, tn)
  call create_basis_function_table(m, y, tn, ty)

end subroutine create_tables


!
! ʻyɸꤹ
! 
! y(i): iܤΥǥåγʻyɸ
!
! , yɸȤ Gauss-Lobatto grid .
!
subroutine set_axis(m, y)

  implicit none

  integer, intent(in)  :: m
  real(8), intent(out) :: y(1:m)  ! ʻyɸ

  integer j
  real(8) pi, dtheta

  pi = acos(-1.0d0)
  dtheta = pi/dble(m+1)
  do j=1,m
     y(j) = cos( dble(j)*dtheta )
  end do

end subroutine set_axis


!
! Chebychev¿༰Υơ֥.
!
subroutine create_chebychev_polynomial_table(m, y, tn)

  implicit none

  integer, intent(in) :: m
  real(8), intent(in) :: y(1:m)
  real(8), intent(out) :: tn(0:4, 1:m, 0:m-1)
  ! tn(k,j,s) y(j) ˤ s Chebychev¿༰ 
  ! k ʬ d^k T_s/dy^k (y(j)) ͤǼ

  integer j, s
  
  ! d^k T_s/dy^k (y)  s=0,1 ˤĤƤͽƤ
  tn(0,1:m,0) = 1.0d0
  tn(0,1:m,1) = y(1:m)
  tn(1,1:m,0) = 0.0d0
  tn(1,1:m,1) = 1.0d0
  tn(2:4,1:m,0:1) = 0.0d0
  do s=2,m-1
     do j=1,m
        ! T_{n+1}(y) - 2yT_n(y) + T_{n-1}(y) = 0Ѥ.
        tn(0,j,s) = 2.0d0*y(j)*tn(0,j,s-1) - tn(0,j,s-2)
        ! ߴط T_{n \pm 1}(y) = ( y \mp (1-y^2)/n d/dy ) T_n(y) Ѥ.
        tn(1,j,s) = dble(s)*( tn(0,j,s-1) - y(j)*tn(0,j,s) ) / (1.0d0-y(j)**2)
        ! ӥդʬ (1-y^2) T_n'' - yT_n' + n^2T_n = 0Ѥ.
        tn(2,j,s) = ( y(j)*tn(1,j,s) - dble(s)**2*tn(0,j,s) ) / (1.0d0-y(j)**2)
        ! ӥդʬyʬΤѤ.
        tn(3,j,s) = ( 3.0d0*y(j)*tn(2,j,s) - (dble(s)**2 - 1.0d0)*tn(1,j,s) ) &
             &                 / (1.0d0-y(j)**2)
        ! ӥդʬy2ʬΤѤ.
        tn(4,j,s) = ( 5.0d0*y(j)*tn(3,j,s) - (dble(s)**2 - 4.0d0)*tn(2,j,s) ) &
             &                 / (1.0d0-y(j)**2)
     end do
  end do

end subroutine create_chebychev_polynomial_table


!
! Chebychev¿༰ѤyŸѤؿɽ.
! 
! ؿ F_n(y) := (1-y^2)^2 T_n(y) Ȥ.
! d^k F_n/dy^k (y)ͤk=04ޤǵ.
!
! Note: δؿ, Boyd(2001)p.144Heinrichs' basisȤƺܤäƤ.
!
subroutine create_basis_function_table(m, y, tn, ty)

  implicit none

  integer, intent(in) :: m
  real(8), intent(in) :: y(1:m), tn(0:4, 1:m, 0:m-1)
  real(8), intent(out) :: ty(0:4, 1:m, 0:m-1) ! ؿ

  integer j, s
  
  ! s δؿ k ʬ y(j) Ǥ d^k F_s/dy^k (y(j)) ɽ ty(k,j,s) 
  ! ʲδطϴؿñykʬƳФǤ. 
  do s=0,m-1
     do j=1,m
        ty(0,j,s) = (1.0d0-y(j)**2)**2*tn(0,j,s)

        ty(1,j,s) = -4.0d0*y(j)*(1.0d0-y(j)**2)*tn(0,j,s) &
             &      +(1.0d0-y(j)**2)**2*tn(1,j,s)
        
        ty(2,j,s) = (-4.0d0+12.0d0*y(j)**2)*tn(0,j,s) &
             &      -8.0d0*y(j)*(1.0d0-y(j)**2)*tn(1,j,s) &
             &      +(1.0d0-y(j)**2)**2*tn(2,j,s)

        ty(3,j,s) = 24.0d0*y(j)*tn(0,j,s) &
             &      +(-12.0d0+36.0d0*y(j)**2)*tn(1,j,s) &
             &      -12.0d0*y(j)*(1.0d0-y(j)**2)*tn(2,j,s) &
             &      +(1.0d0-y(j)**2)**2*tn(3,j,s)

        ty(4,j,s) = 24.0d0*tn(0,j,s) &
             &      +96.0d0*y(j)*tn(1,j,s) &
             &      +(-24.0d0+72.0d0*y(j)**2)*tn(2,j,s) &
             &      -16.0d0*y(j)*(1.0d0-y(j)**2)*tn(3,j,s) &
                    +(1.0d0-y(j)**2)**2*tn(4,j,s)
     end do
  end do
  
end subroutine create_basis_function_table


!
! ʿ̥ݥ梁ήδܾꤹ
!
subroutine set_meanflow(m, y, u, d2u)
  
  implicit none
  
  integer, intent(in)  :: m         ! yΥǥåκ
  real(8), intent(in)  :: y(1:m)    ! ʻ y ɸ
  real(8), intent(out) :: u(1:m), & ! ή®, $ u = 1-y**2 $
       &                  d2u(1:m)  ! uy˴ؤ2ʬ, $ d^2 u/dy^2 = -2 $

  u(1:m) = 1.0d0 - y(1:m)**2
  d2u(1:m) = -2.0d0
  
end subroutine set_meanflow


! 
! Orr-SommerfeldβؿǵŸȤŸˤĤƤ
! ̲ͭ $A a = \sigma B a$ ˸ A, B׻
! 
!  a  ؿ F_n(y(iy))η a_n (1 m), 
! A,Bj,iǤca(i,j),cb(i,j)˳Ǽ
! 
! 
subroutine calculate_matrices(m, ty, u, d2u, re, al, ca, cb)

  implicit none

  integer, intent(in) :: m
  real(8), intent(in) :: ty(0:4, 1:m, 0:m-1), u(1:m), d2u(1:m), re, al
  complex(8), intent(out) :: ca(1:m, 1:m), cb(1:m, 1:m)

  integer i, j

  ca(1:m, 1:m) = 0.0d0
  cb(1:m, 1:m) = 0.0d0
  do j=0,m-1
     do i=1,m
        ca(i,j+1) = u(i)*(ty(2,i,j) - al**2*ty(0,i,j)) &
             &      -d2u(i)*ty(0,i,j) &
             &      +dcmplx(0.0d0, 1.0d0)*( ty(4,i,j) - 2.0d0*al**2*ty(2,i,j) + al**4*ty(0,i,j) )/(al*re)
        cb(i,j+1) = ty(2,i,j) - al**2*ty(0,i,j)
     end do
  end do

end subroutine calculate_matrices


!
! ޤäͭ٥ȥ뤫, ήؿ԰⡼/Ǿ⡼ɤ
! 
subroutine calculate_eigenmode(m, y, ty, eigenvector)

  implicit none

  integer, intent(in)  :: m
  real(8), intent(in)  :: y(1:m), ty(0:4, 1:m, 0:m-1)
  complex(8), intent(in) :: eigenvector(1:m)

  complex(8) em(0:m+1), a
  real(8) y0(0:m+1) ! y(1:m)˶κɸդä
  integer j, s

  ! ޤ᤿yɸ y0 
  y0(0) = 1.0d0
  y0(1:m) = y(1:m)
  y0(m+1) = -1.0d0
  
  em(0) = 0.0d0   ! ؿ϶Ǿ˥ʤΤ, ͭ⡼ɤ⥼
  em(m+1) = 0.0d0 ! 
  em(1:m) = 0.0d0
  do j=1,m
     do s=1,m
        em(j) = em(j) + eigenvector(s)*ty(0,j,s-1)
     end do
  end do
  a = em(m/2+1) ! y=0Ǥʣǿ(ʲ˻Ȥ)

100 format(3(1x,1pe14.7))
  do j=0,m+1
     write(*, 100) y0(j), em(j)/a
  end do

end subroutine calculate_eigenmode


! 
! ͭ, ԰⡼ɤޤϺǾ⡼ɤ, б
! ͭؿ
! 
! Note:
!    оιΰ̲ͭLapackΥ֥롼 zgegv Ѥ.
!
!    zgegvλȤ, 
!    call zgegv( 'N', 'V', N, A, LDA, B, LDB, ALPHA, BETA,
!                VL, LDVL, VR, LDVR, WORK, LWORK, RWORK, INFO )
!
!    LDA, LDB leading dimension of A, B
!    ͭ lambda  lambda = alpha/beta η֤.
!    VLϺͭ٥ȥ, VRϱͭ٥ȥ
!    ܤȤ http://www.netlib.org/lapack/complex16/zgegv.f 򻲾Ȥ뤳.
!
subroutine solve_eigenvalue_problem(m, ca, cb, eigenvector, eigenvalue)

  implicit none

  integer, intent(in) :: m
  complex(8), intent(inout) :: ca(1:m, 1:m), cb(1:m, 1:m)
  complex(8), intent(out) :: eigenvector(1:m), eigenvalue

  integer info
  complex(8) alpha(1:m), beta(1:m), dummy(1:1, 1:m), rev(1:m, 1:m), work(1:24*m)
  real(8) work2(1:8*m)
  
  complex(8) lambda(1:m)
  integer i_max

  call zgegv( 'N', 'V', m, ca, m, cb, m, alpha, beta, &
       &      dummy, 1, rev, m, work, 24*m, work2, info )
  if( info /= 0 )then
     print *, 'error has occured in zgegv: error_code=', info
     call exit(1)
  end if
  lambda(1:m) = alpha(1:m)/beta(1:m)
  
  i_max = maxloc(dimag(lambda), 1)

  eigenvector(1:m) = rev(1:m, i_max)
  eigenvalue = lambda(i_max)

end subroutine solve_eigenvalue_problem
